using System;
using System.Threading;
using System.IO;
using System.Reflection.Emit;
using System.Reflection;
using System.Collections.Generic;
using System.Text;

namespace BFNet
{
    class Program
    {

        private void GenerateIL(string fileName, ILGenerator il)
        {
            string file = File.OpenText(fileName).ReadToEnd();
            Stack<Label> jumpForwardTable = new Stack<Label>();
            Stack<Label> jumpBackTable = new Stack<Label>();
            // Todo: 16 bit vs. 8 bit BF
            il.DeclareLocal(typeof(char[]));
            il.DeclareLocal(typeof(int));
            il.DeclareLocal(typeof(ConsoleKeyInfo));
            // Todo: Buffer size issues
            il.Emit(OpCodes.Ldc_I4, 30000);
            il.Emit(OpCodes.Newarr, typeof(char));
            il.Emit(OpCodes.Stloc_0);
            il.Emit(OpCodes.Ldc_I4, 0);
            il.Emit(OpCodes.Stloc_1);            

            for (int i = 0; i < file.Length; i++)
            {
                switch (file[i])
                {
                    case '>':
                        {
                            il.Emit(OpCodes.Ldloc_1);

                            int total = 1;

                            while (i + 1 < file.Length && file[i + 1] == '>')
                            {
                                i++;
                                total++;
                            }

                            il.Emit(OpCodes.Ldc_I4, total);
                            il.Emit(OpCodes.Add);
                            il.Emit(OpCodes.Stloc_1);
                            break;
                        }
                    case '<':
                        {
                            il.Emit(OpCodes.Ldloc_1);

                            int total = 1;

                            while (i + 1 < file.Length && file[i + 1] == '<')
                            {
                                i++;
                                total++;
                            }

                            il.Emit(OpCodes.Ldc_I4, total);
                            il.Emit(OpCodes.Sub);
                            il.Emit(OpCodes.Stloc_1);
                            break;
                        }
                    case '+':
                        {
                            il.Emit(OpCodes.Ldloc_0);
                            il.Emit(OpCodes.Ldloc_1);

                            il.Emit(OpCodes.Ldloc_0);
                            il.Emit(OpCodes.Ldloc_1);

                            il.Emit(OpCodes.Ldelem_I1);

                            int total = 1;

                            while (i + 1 < file.Length && file[i + 1] == '+')
                            {
                                i++;
                                total++;
                            }

                            il.Emit(OpCodes.Ldc_I4, total);


                            il.Emit(OpCodes.Add);
                            il.Emit(OpCodes.Stelem_I1);
                            break;
                        }
                    case '-':
                        {
                            il.Emit(OpCodes.Ldloc_0);
                            il.Emit(OpCodes.Ldloc_1);

                            il.Emit(OpCodes.Ldloc_0);
                            il.Emit(OpCodes.Ldloc_1);

                            il.Emit(OpCodes.Ldelem_I1);

                            int total = 1;

                            while (i + 1 < file.Length && file[i + 1] == '-')
                            {
                                i++;
                                total++;
                            }

                            il.Emit(OpCodes.Ldc_I4, total);
                            il.Emit(OpCodes.Sub);
                            il.Emit(OpCodes.Stelem_I1);
                            break;
                        }
                    case '[':
                        {
                            jumpBackTable.Push(il.DefineLabel());
                            jumpForwardTable.Push(il.DefineLabel());
                            
                            il.Emit(OpCodes.Br, jumpForwardTable.Peek());
                            il.MarkLabel(jumpBackTable.Peek());

                            /*
                            jumpBackTable.Push(il.DefineLabel());
                            jumpForwardTable.Push(il.DefineLabel());

                            il.MarkLabel(jumpBackTable.Peek());
                            il.Emit(OpCodes.Ldc_I4, 0);
                            il.Emit(OpCodes.Ldloc_0);
                            il.Emit(OpCodes.Ldloc_1);
                            il.Emit(OpCodes.Ldelem_I1);
                            il.Emit(OpCodes.Ceq);
                            il.Emit(OpCodes.Brtrue, jumpForwardTable.Peek());
                            */
                            break;
                        }
                    case ']':
                        {

                            il.MarkLabel(jumpForwardTable.Peek());
                            il.Emit(OpCodes.Ldc_I4, 0);
                            il.Emit(OpCodes.Ldloc_0);
                            il.Emit(OpCodes.Ldloc_1);
                            il.Emit(OpCodes.Ldelem_I1);
                            il.Emit(OpCodes.Ceq);
                            il.Emit(OpCodes.Brfalse, jumpBackTable.Peek());

                            jumpBackTable.Pop();
                            jumpForwardTable.Pop();

                            /*
                            il.Emit(OpCodes.Br, jumpBackTable.Peek());
                            il.MarkLabel(jumpForwardTable.Peek());

                            jumpBackTable.Pop();
                            jumpForwardTable.Pop();
                             */
                            break;
                        }
                    case '.':
                        {
                            il.Emit(OpCodes.Ldloc_0);
                            il.Emit(OpCodes.Ldloc_1);
                            il.Emit(OpCodes.Ldelem_I1);
                            //il.Emit(OpCodes.Ldc_I4, 'a');
                            il.Emit(OpCodes.Call, typeof(Console).GetMethod("Write", new Type[1] { typeof(char) }));
                            break;
                        }
                    case ',':
                        {
                            il.Emit(OpCodes.Ldloc_0);
                            il.Emit(OpCodes.Ldloc_1);
                            il.Emit(OpCodes.Call, typeof(Console).GetMethod("ReadKey", new Type[0]));
                            il.Emit(OpCodes.Stloc_2);
                            il.Emit(OpCodes.Ldloca_S, 2); 
                            il.Emit(OpCodes.Call, typeof(ConsoleKeyInfo).GetMethod("get_KeyChar", new Type[0]));
                            il.Emit(OpCodes.Stelem_I1);
                            
                            break;
                        }

                }
                                
            }


            il.Emit(OpCodes.Ret);
        }

        private void Build(string fileName)
        {
            string baseName = fileName.Split('.')[0];
            AppDomain myDomain = Thread.GetDomain();
            AssemblyName asmName = new AssemblyName(baseName + "_assembly");
            AssemblyBuilder asmBuilder = myDomain.DefineDynamicAssembly(asmName, AssemblyBuilderAccess.RunAndSave);
            ModuleBuilder bfModule = asmBuilder.DefineDynamicModule(baseName + "_module", baseName + ".exe");
            TypeBuilder typBuilder = bfModule.DefineType("BFMain", TypeAttributes.Public);

            Type objType = Type.GetType("System.Object");
            ConstructorInfo objCtor = objType.GetConstructor(new Type[0]);
            ConstructorBuilder bfCtor = typBuilder.DefineConstructor(MethodAttributes.Public, CallingConventions.Standard, new Type[0]);
            ILGenerator ctorIL = bfCtor.GetILGenerator();

            
            ctorIL.Emit(OpCodes.Ldarg_0);
            ctorIL.Emit(OpCodes.Call, objCtor);
            ctorIL.Emit(OpCodes.Ret);

            MethodBuilder mainMethod = typBuilder.DefineMethod("Main", MethodAttributes.Public | MethodAttributes.Static, typeof(void), new Type[] { typeof(string[]) });

            ILGenerator mainIL = mainMethod.GetILGenerator();

            GenerateIL(fileName, mainIL);
            
            typBuilder.CreateType();
            asmBuilder.SetEntryPoint(typBuilder.GetMethod("Main"));
            asmBuilder.Save(baseName + ".exe");
        }


        // Takes a filename as a parameter, outputs an exe
        static void Main(string[] args)
        {
            //Console.Write(Console.ReadKey().KeyChar);
            if (args.Length == 1 && File.Exists(args[0]))
            {

                new Program().Build(args[0]);
            }
            else
            {
                new Program().Build("99botles.bf");
                //Console.WriteLine("Invalid arguments: Expecting a single argument containing filename of bf file to compile");
            }
        }
    }
}
